; =============================================================================
;	Horizontal scroll sample program for MSX2
; -----------------------------------------------------------------------------
;	2019/8/10	t.hara
; =============================================================================

	include		"msx.asm"

	bsave_header	initialize, end_address, initialize

; =============================================================================
;	initialize
; =============================================================================
	org			0xB000
initialize:
	call		sprite_initializer

; =============================================================================
;	main routine
; =============================================================================
main:
	call		wait_vsync
	call		sprite_move
	jp			main

; =============================================================================
;	h_timi_interrupt handler
;	input)
;		none
;	output)
;		none
;	break)
;		all
; =============================================================================
	scope		h_timi_interrupt_handler
h_timi_interrupt_handler::
	ld			a, 1
	ld			[vsync_flag], a
	call		horizontal_scroll
h_timi_next::
	ret
	ret
	ret
	ret
	ret

vsync_flag::
	db			0
	endscope

; =============================================================================
;	wait vsync
;	input)
;		none
;	output)
;		none
; =============================================================================
	scope		wait_vsync
wait_vsync::
	halt
	ld			a, [vsync_flag]
	or			a, a
	jp			z, wait_vsync
	
	xor			a, a
	ld			[vsync_flag], a
	ret
	endscope

; =============================================================================
;	horizontal scroll
;	input)
;		a ... current status number
;	output)
;		none
; =============================================================================
	scope		horizontal_scroll
horizontal_scroll::
	call		wait_vdp_command

	ld			a, [phase]
	inc			a
	and			a, 15
	ld			[phase], a
	jp			nz, skip1

	ld			a, [draw_page]
	xor			a, 1
	ld			[draw_page], a
	xor			a, 1					;	A is display page number.
	ld			a, 0b0001_1111
	jp			z, skip0
	ld			a, 0b0011_1111
skip0:
	out			[IO_VDP_PORT1], a
	ld			a, 0x80 | 2
	out			[IO_VDP_PORT1], a
	xor			a, a					;	A is 0. It is phase number.
skip1:
	;	update sprite_add_num
	ld			[sprite_add_num], a		;	sprite_add_num = phase xor 15

	;	R#18 = A xor 8
	ld			b, a
	xor			a, 8
	out			[IO_VDP_PORT1], a
	ld			a, 0x80 | 18
	out			[IO_VDP_PORT1], a
	ld			a, b

	;	branch phases
	cp			a, 8
	jp			c, phase_0_7
	and			a, 7
	cp			a, 6
	ret			nc

	;	case of phase 8...13
phase_8_13:
	; --------------------------------------
	;	Copy parts P to N

	;	dy = N << 5
	rrca										;	note) 0b0000_0NNN --> 0bNNN0_0000
	rrca
	rrca
	ld			[vdp_command_3a_dest_num], a
	;	get random (P)
	call		random
	and			a, 1							;	note) Since there are only two types of prepared parts, the value is limited
	ld			b, a
	;	sx = (P and 15) << 4
	and			a, 15
	rlca
	rlca
	rlca
	rlca
	ld			[vdp_command_3a_part_num_x], a
	;	sy = P and 0xF0
	ld			a, 0xF0
	and			a, b
	ld			[vdp_command_3a_part_num_y], a
	;	page = draw_page
	ld			a, [draw_page]
	ld			[vdp_command_3a_dest_page], a
	ld			hl, vdp_command_3a
	call		run_vdp_command

	; --------------------------------------
	;	Copy parts P to N+1

	;	dy = dy + 16
	ld			a, [vdp_command_3a_dest_num]
	add			a, 16
	ld			[vdp_command_3a_dest_num], a
	;	get random (P)
	call		random
	and			a, 1							;	note) Since there are only two types of prepared parts, the value is limited
	ld			b, a
	;	sx = (P and 15) << 4
	and			a, 15
	rlca
	rlca
	rlca
	rlca
	ld			[vdp_command_3a_part_num_x], a
	;	sy = P and 0xF0
	ld			a, 0xF0
	and			a, b
	ld			[vdp_command_3a_part_num_y], a
	ld			hl, vdp_command_3a
	call		wait_vdp_command
	call		run_vdp_command
	ret

	;	case of phase 0...7
phase_0_7:
	;	sy, dy = N * 24 = N * 16 + N * 8
	rlca
	rlca
	rlca
	ld			b, a
	rlca
	add			a, b
	ld			[vdp_command_1a + 2], a
	ld			[vdp_command_1a + 6], a
	ld			[vdp_command_1b + 2], a
	ld			[vdp_command_1b + 6], a

	ld			a, [draw_page]
	or			a, a
	ld			hl, vdp_command_1a
	jp			nz, skip2
	ld			hl, vdp_command_1b
skip2:
	call		run_vdp_command
	ret

sprite_add_num::
	db			0
phase:
	db			15		; This is program phase. This range is 0~15.
draw_page::
	db			0
	endscope

; =============================================================================
;	sprite initializer
;	input)
;		none
;	output)
;		none
;	break)
;		all
; =============================================================================
	scope		sprite_initializer
sprite_initializer::

	;	R#9 = R#9 and 0b0111_1111 (192line mode)
	ld		a, [reg9sav]
	and		a, 0b0111_1111
	out		[IO_VDP_PORT1], a
	ld		a, 0x80 | 9
	out		[IO_VDP_PORT1], a

	;	initialize sprite generator table for Sprite pattern #0, #1
	ld			hl, sprite_pattern
	ld			de, SC5_SPR_GEN
	ld			bc, 64
	call		ldirvm

	;	initialize sprite color table for Sprite #0, #1
	ld			hl, sprite_color
	ld			de, SC5_SPR_COL
	ld			bc, 32
	call		ldirvm

	di
	;	h_timi
	ld		hl, h_timi			; Source address
	ld		de, h_timi_next		; Destination address
	ld		bc, 5				; Transfer length
	ldir						; Block transfer

	ld		hl, h_timi_new		; Source address
	ld		de, h_timi			; Destination address
	ld		bc, 3				; Transfer length
	ldir						; Block transfer
	ei
	ret
h_timi_new:
	jp		h_timi_interrupt_handler	; Jump to h_timi_interrupt_handler. This is 3bytes code.
	endscope

; =============================================================================
;	sprite move
;	input)
;		none
;	output)
;		none
;	break)
;		all
; =============================================================================
	scope		sprite_move
sprite_move::
	;	get stick value
	xor			a, a
	call		gtstick		; get stick(0)
	push		af
	ld			a, 1
	call		gtstick		; get stick(1)
	pop			bc
	or			a, b
	ld			c, a
	ld			b, 0
	ld			hl, direction_y_table
	adc			hl, bc
	ld			c, [hl]		; b = direction_y_table[ stick(0) or stick(1) ]
	ld			a, [player_y]
	add			a, c

	cp			a, -2
	jp			nz, skip0
	xor			a, a
skip0:
	cp			a, 192 - 16 + 2
	jp			nz, skip1
	ld			a, 192 - 16
skip1:
	ld			[player_y], a
	ld			[sprite_attribute + 0], a
	ld			[sprite_attribute + 4], a

	ld			c, 9
	add			hl, bc
	ld			c, [hl]		; b = direction_x_table[ stick(0) or stick(1) ]
	ld			a, [player_x]
	add			a, c

	cp			a, -2
	jp			nz, skip2
	xor			a, a
skip2:
	cp			a, 256 - 16 + 2
	jp			nz, skip3
	ld			a, 256 - 16
skip3:
	ld			[player_x], a

	ld			b, a
	ld			a, [sprite_add_num]
	add			a, b
	ld			[sprite_attribute + 1], a
	ld			[sprite_attribute + 5], a

	ld			hl, sprite_attribute
	ld			de, SC5_SPR_ATTR
	ld			bc, 8
	call		ldirvm
	ret
direction_y_table:
	db			0			; 0
	db			-2			; 1
	db			-2 			; 2
	db			0 			; 3
	db			2 			; 4
	db			2 			; 5
	db			2 			; 6
	db			0			; 7
	db			-2			; 8
direction_x_table:
	db			0			; 0
	db			0			; 1
	db			2			; 2
	db			2			; 3
	db			2			; 4
	db			0			; 5
	db			-2			; 6
	db			-2			; 7
	db			-2			; 8
player_x:
	db			128-8
player_y:
	db			96-8
	endscope

; =============================================================================
;	random
;	input)
;		none
;	output)
;		a ... random value 0...255
;	break)
;		af, b
; =============================================================================
	scope		random
random::
	ld			a, [random_seed1]
	rlca
	ld			b, a
	ld			a, [random_seed2]
	rrca
	rrca
	xor			a, b
	dec			a
	ld			[random_seed1], a
	ld			a, b
	inc			a
	ld			[random_seed2], a
	ret
random_seed1:
	db			0b1001_1101
random_seed2:
	db			0b1010_0011
	endscope

; =============================================================================
;	run vdp command
;	input)
;		hl ... register table address (VDP R#32...R#46)
; =============================================================================
	scope		run_vdp_command
run_vdp_command::
	;	VDP R#17 = 32
	ld			a, 32
	out			[IO_VDP_PORT1], a
	ld			a, 0x80 | 17
	out			[IO_VDP_PORT1], a
	;	R#32~46 (15 registers)
	ld			bc, (15 << 8) | IO_VDP_PORT3
	otir
	ret
	endscope

; =============================================================================
;	wait vdp command
; =============================================================================
	scope		wait_vdp_command
wait_vdp_command::
	;	VDP R#15 = 2
	ld			a, 2
	ld			bc, ((0x80 | 15) << 8) | IO_VDP_PORT1
	out			[c], a
	out			[c], b
loop:
	in			a, [c]
	and			a, 1
	jp			nz, loop
	xor			a, a
	out			[c], a
	out			[c], b
	ret
	endscope

; =============================================================================
;	[1]-(a) VDP Command HMMM (Page0 to Page1)
; =============================================================================
vdp_command_1a:
	db			16, 0, 0, 0, 0, 0, 0, 1, 240, 0, 24, 0, 0, 0, 0b11010000

; =============================================================================
;	[1]-(b) VDP Command HMMM (Page1 to Page0)
; =============================================================================
vdp_command_1b:
	db			16, 0, 0, 1, 0, 0, 0, 0, 240, 0, 24, 0, 0, 0, 0b11010000

; =============================================================================
;	[3]-(a)(b) VDP Command HMMM (Page2 to Page1/0)
; =============================================================================
vdp_command_3a:
vdp_command_3a_part_num_x:
	db			0			;	R#32 sx
	db			0			;	R#33 sx
vdp_command_3a_part_num_y:
	db			0			;	R#34 sy
	db			2			;	R#35 sy
	db			240			;	R#36 dx
	db			0			;	R#37 dx
vdp_command_3a_dest_num:
	db			0			;	R#38 dy
vdp_command_3a_dest_page:
	db			1			;	R#39 dy
	db			16, 0, 16, 0, 0, 0, 0b11010000		;	R#40, 41, 42, 43, 44, 45, 46

; =============================================================================
;	Sprite pattern
; =============================================================================
sprite_pattern:
	db			0x00, 0x00, 0x70, 0x78, 0x7c, 0x7d, 0x7e, 0x7f
	db			0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 0x7e, 0x78
	db			0x00, 0x00, 0x00, 0x00, 0x70, 0x88, 0x04, 0x02
	db			0x82, 0xff, 0xff, 0xff, 0xfe, 0x80, 0x00, 0x00

	db			0x00, 0x00, 0x00, 0x30, 0x38, 0x38, 0x3d, 0x3e
	db			0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x3e, 0x38, 0x00
	db			0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 0xf8, 0xfc
	db			0x7c, 0x82, 0xfe, 0xfe, 0x80, 0x00, 0x00, 0x00

; =============================================================================
;	Sprite color
; =============================================================================
sprite_color:
	db			0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
	db			0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01

	db			0x4e, 0x4e, 0x4e, 0x4e, 0x4e, 0x4e, 0x4e, 0x4e
	db			0x4e, 0x4e, 0x4e, 0x4e, 0x4e, 0x4e, 0x4e, 0x4e

; =============================================================================
;	Sprite attribute
; =============================================================================
sprite_attribute:
	db			0		; Y
	db			0		; X
	db			0		; Pattern#
	db			0		; N/A

	db			0		; Y
	db			0		; X
	db			4		; Pattern#
	db			0		; N/A
end_address:
